use serde::{Serialize, Deserialize};

use super::{order, Priority, Service, ServiceAPI};
use crate::io::Packet;

use crate as fleck;
use crate::msg::MessageParams;

crate::service_priority!(
    ParsePacket,
    in: Priority<order::First, order::Preprocessing>
);
#[derive(Default)]
pub(crate) struct ParsePacket {}

impl Service for ParsePacket {
    fn process_incoming(&mut self, api: &dyn ServiceAPI, _packet: &mut Packet) {
        // try deserializing
        // get message registry from api
        // remove data from packet, add parsed Msg instead
        println!("incoming packet!");
    }
}

crate::service_priority!(
    SendPacket,
    out: Priority<order::Postprocessing, order::Last>
);
#[derive(Default)]
pub(crate) struct SendPacket {}

impl Service for SendPacket {
    fn process_outgoing(&mut self, api: &dyn ServiceAPI, packet: &mut crate::io::Packet) {
        // use default channel if not specified
        if packet.io_channel.is_none() {
            packet.io_channel = Some(api.raw_io().global().clone());
        }

        // serialize if needed
        if let (false, Some(msg)) = (packet.data.as_ref().is_some(), packet.msg.take()) {
            packet.data = Some(bincode::serialize(&msg).expect("failed to serialize message"));
        }

        match &mut packet.io_channel {
            Some(ch) => ch.clone().send_packet(packet),
            _ => {
                println!("tried to send packet with no channel?");
            }
        }
    }
}

crate::service_priority!(LocalDiscovery, out: Priority<order::Processing, order::Postprocessing>);
#[derive(Default)]
pub(crate) struct LocalDiscovery;


#[derive(Serialize, Deserialize)]
struct DiscoveryMsg {
    pkey: crate::crypto::PublicKey
}

impl MessageParams for DiscoveryMsg {
    const NAME: &'static str = "DiscoveryMsg";
    const ENCRYPTED: bool = false;
}

impl Service for LocalDiscovery {
    fn register_msgs(&self, registry: &mut fleck::msg::MessageRegistry) {
        registry.add_message_type::<DiscoveryMsg>();
    }

    fn process_incoming(&mut self, api: &dyn ServiceAPI, packet: &mut Packet) {

    }

    fn process_major_tick(&mut self, api: &dyn ServiceAPI) {
        api.services().with_service(|n: &super::nodes::Nodes| {
            
        });

        api.send_packet(Packet {
            addr: Some((*crate::io::MULTICAST_ADDRESS).into()),
            data: None,
            io_channel: Some(api.raw_io().local().clone()),
            msg: Some(crate::msg::Message::build(DiscoveryMsg { pkey: Default::default() })),
        });
    }
}
